import numpy as np

from ..utils import inside_box_coords, cs2span, listify, list2str
from ..utils import log_and_raise
from ..utils.check import check_3D_lists
from ..constants import int_, float_, fp_eps, xyz_list, xyz_dict, C_0

class Monitor(object):
    """Base class for defining field monitors.
    """
    def __init__(self, center, size, store, name=None):
        """Base constructor. Available subclasses:

        - :class:`.TimeMonitor`
        - :class:`.FreqMonitor`
        - :class:`.ModeMonitor`
        """
        check_3D_lists(center=listify(center), size=listify(size))
        self.center = np.array(center)
        self.size = np.array(size)
        """ Slightly offset the span such that if the monitor is zero-size 
        along a certain direction, and is placed exactly on a grid coordinate,
        it will deterministically be placed to the right of the coordinate."""
        self.span = cs2span(self.center, self.size)
        zero_span = self.span[:, 1] - self.span[:, 0] < fp_eps
        self.span[zero_span, :] += fp_eps
        self.store = []
        for f in listify(store):
            if f.lower() in ['e', 'h', 'flux']:
                self.store.append(f.lower())
            else:
                log_and_raise(
                    f"Unrecognized store value {f}. Valid values are "
                    "'E', 'H', and 'flux'.",
                    ValueError
                )

        self.name = None if name is None else str(name)


    def _inside(self, coords):
        """ Get a mask equal to one if a point is inside the monitor region, 
        and zero if outside.
        
        Parameters
        ----------
        coords : 3-tuple
            Defines the x, y, and z coords. 
        """
        mask = np.zeros(tuple(c.size - 1 for c in coords))
        indsx, indsy, indsz = inside_box_coords(self.span, coords)
        mask[indsx[0]:indsx[1], indsy[0]:indsy[1], indsz[0]:indsz[1]] = 1.0

        return mask

    def _inside_inds(self, coords):
        """ Get indexes of the points inside the monitor region.
        
        Parameters
        ----------
        coords : 3-tuple
            Defines the x, y, and z coords. 
        
        Returns
        -------
        np.ndarray
            An array of shape (Np, 3), where Np is the total number of coords 
            points in the monitor region.
        """
        inds = inside_box_coords(self.span, coords)
        indsm = np.meshgrid(np.arange(inds[0][0], inds[0][1]), 
                            np.arange(inds[1][0], inds[1][1]),
                            np.arange(inds[2][0], inds[2][1]),
                            indexing='ij')
        if indsm[0].size==0: 
            return np.zeros((0, 3), dtype=int_)

        return np.stack([inds.ravel() for inds in indsm], axis=1).astype(int_)


class TimeMonitor(Monitor):
    """Monitor recording the time-domain fields within a 3D region.
    """

    def __init__(self, center, size, t_start=0, t_stop=None, t_step=None,
                    store=('E', 'H'), name=None):
        """Construct.
        
        Parameters
        ----------
        center : array_like
            (micron) x, y, and z position of the center of the Monitor.
        size : array_like
            (micron) Size in x, y, and z.
        t_start : float, optional
            (second) Starting time of field recording.
        t_stop : float, optional
            (second) Stopping time of field recording. If ``None``, record 
            until the end of the simulation.
        t_step : None, optional
            (second) : Time interval at which to record the fields. If
            ``None``, the fields at every time step are recorded. If a float,
            it is snapped to an integer multiple of the simulation time step.
        store : list, optional
            List of attributes to be recorded. Valid entries are 
            ``'E'``, ``'H'`` and ``'flux'``. If ``'flux'`` is requested, the 
            monitor must be a 2D surface.
        name : str, optional
            Custom name of the monitor.
        
        Note
        ----
        Time monitors can result in very large amounts of data if defined over 
        a large spatial region. Recommended usage is either recording the full 
        time evolution of a single point in space, or using ``t_start`` and 
        ``t_stop`` to record just a few time steps of a larger region. 
        """


        super().__init__(center, size, store, name)
        self.t_start = t_start
        self.t_stop = t_stop
        self.t_step = t_step

    def __repr__(self):
        rep = "Tidy3D TimeMonitor: {\n"
        rep += "name     = %s\n"%self.name
        rep += "center   = %s\n" % list2str(self.center, "%1.4f")
        rep += "size     = %s\n" % list2str(self.size, "%1.4f")
        rep += "t_start  = %1.2e,\n"%self.t_start
        if self.t_stop is None:
            rep += "t_stop   = None\n"
        else:
            rep += "t_stop   = %1.2e,\n"%self.t_stop
        if self.t_step is None:
            rep += "t_step   = None\n"
        else:
            rep += "t_step   = %1.2e,\n"%self.t_step
        rep += f"Store: {self.store}\n"
        rep += "}\n"

        return rep

class FreqMonitor(Monitor):
    """Monitor recording a discrete Fourier transform of the fields within a 
    3D region, for a given list of frequencies.
    """
    
    def __init__(self, center, size, freqs, store=('E', 'H'), name=None):
        """ Construct.

        Parameters
        ----------
        center : array_like
            (micron) x, y, and z position of the center of the Monitor.
        size : array_like
            (micron) Size in x, y, and z.
        freqs : float or array_like
            Frequencies at which the fields are sampled.
        store : list, optional
            List of attributes to be recorded. Valid entries are 
            ``'E'``, ``'H'`` and ``'flux'``. If ``'flux'`` is requested, the 
            monitor must be a 2D surface.
        name : str, optional
            Custom name of the monitor.
        """

        super().__init__(center, size, store, name)
        self.freqs = listify(freqs)
        if len(self.freqs) == 0:
            log_and_raise(
                "Monitor 'freqs' must be a non-empty list.", ValueError
            )
        self.lambdas = C_0 / np.array(freqs)

    def __repr__(self):
        rep = "Tidy3D FreqMonitor: {\n"
        rep += "name     = %s\n"%self.name
        rep += "center   = %s\n" % list2str(self.center, "%1.4f")
        rep += "size     = %s\n" % list2str(self.size, "%1.4f")
        rep += "freqs    = %s\n" % list2str(self.freqs, "%1.2e")
        rep += f"Store: {self.store}\n"
        rep += "}\n"

        return rep

class ModeMonitor(FreqMonitor):
    """ :class:`.FreqMonitor` subclass defining a 2D plane in which the 
    recorded frequency-domain fields can be decomposed into propagating 
    eigenmodes.
    """
    
    def __init__(self, center, size, freqs, Nmodes=1, name=None):
        """ Construct.

        Parameters
        ----------
        center : array_like
            (micron) 3D vector defining the center of the 2D plane.
        size : array_like
            (micron) 3D vector defining the size of the 2D plane. Exactly one 
            of the values must be ``0``, defining the normal direction.
        freqs : float or list of float
            Frequencies at which the fields are sampled.
        Nmodes : int
            Compute the decomposition into the first ``Nmodes`` modes in
            order of decreasing effective index.
        name : str, optional
            Custom name of the monitor.

        Note
        ----
        After a simulation run, the frequency-domain E and H fields are always
        stored. In addition, the expansion coefficients for positive and
        negative direction of propagation for the ``Nmodes`` modes are stored.
        """

        super().__init__(center, size, freqs, store=['E', 'H'], name=name)
        self.Nmodes = Nmodes

    def __repr__(self):
        rep = "Tidy3D ModeMonitor: {\n"
        rep += "name     = %s\n"%self.name
        rep += "center   = %s\n" % list2str(self.center, "%1.4f")
        rep += "size     = %s\n" % list2str(self.size, "%1.4f")
        rep += "freqs    = %s\n" % list2str(self.freqs, "%1.2e")
        rep += "Nmodes   = %d\n" % self.Nmodes
        rep += "}\n"

        return rep