import numpy as np

from .Monitor import TimeMonitor, FreqMonitor, ModeMonitor
from ..constants import int_, float_, complex_, fp_eps, C_0
from ..mode import ModePlane, Mode
from ..utils.log import log_and_raise

class MonitorData(object):
    """Class used internally to store data related to a Monitor added in a 
    specific Simulation.
    """

    def __init__(self, monitor):

        # Reference to the monitor that the data corresponds to
        self.monitor = monitor

        # To be set by Simulation
        self.name = None

        # Everything below is set after a run of store_data() or load_fields()
        self.data = False # To be set to True after data is loaded

        # Raw E and H data as returned by the solver
        self._E = np.empty((3, 0, 0, 0, 0), dtype=float_)
        self._H = np.empty((3, 0, 0, 0, 0), dtype=float_)
        # Total flux as returned by the solver
        self._flux = np.empty((0, 1), dtype=float_)
        # Poynting flux density if computed after a solve
        self._S = np.empty((3, 0, 0, 0, 0), dtype=float_)

        # Mesh defining the positions at which the fields are returned
        self.xmesh = np.empty((0, ), dtype=float_)
        self.ymesh = np.empty((0, ), dtype=float_)
        self.zmesh = np.empty((0, ), dtype=float_)
        self.mesh_step = np.empty((3, ), dtype=float_)

        # Mesh in time, for a TimeMonitor. Set when added to a Simulation.
        self.tmesh = np.empty((0, ), dtype=float_)
        self.tind_beg = 0
        self.tind_end = 0
        self.tind_step = 1
        self.Nt = 0

        # Frequencies, for a FreqMonitor
        self.freqs = np.empty((0, ), dtype=float_)
        if isinstance(monitor, FreqMonitor):
            self.freqs = np.array(monitor.freqs, dtype=float_)
            self.Nf = self.freqs.size

        # Indexes defining the span of the Monitor in the simulation grid in 
        # which it is embedded.
        self.mnt_inds = np.empty((0, 3), dtype=int_)
        self.inds_beg = np.zeros((3, ), dtype=int_)
        self.inds_end = np.zeros((3, ), dtype=int_)

        # Source normalization, which may be set later for FreqMonitors
        self.set_source_norm(None)

        # For 2D monitors
        self.norm_ind = None
        self.normal = None

        # Expansion coefficients for a ModeMonitor, loaded after solver run
        self._mode_amps = np.empty((2, 0, 0), dtype=float_)
        if isinstance(monitor, ModeMonitor):
            self.Nmodes = monitor.Nmodes
            self.target_neff = None
            self._mode_amps = np.empty((2, self.Nf, self.Nmodes), dtype=float_)

    @property
    def E(self):
        """ (V/micron) Electric field, if it was requested and after the data
        has been loaded.
        """
        return self._E

    @property
    def H(self):
        """ (A/micron) Magnetic field, if it was requested and after the data
        has been loaded.
        """
        return self._H

    @property
    def S(self):
        """ (W/micron\\ :sup:`2`) Poynting vector (power flux density) if it
        has been computed with :meth:`Simulation.poynting`.
        """
        return self._S

    @property
    def flux(self):
        """ (W) Power flux through a 2D surface monitor, if it was requested
        and after the data has been loaded.
        """
        return self._flux

    @property
    def mode_amps(self):
        """ (:math:`\\sqrt{W}`) Power amplitude expansion coefficients for
        mode monitors.
        """
        return self._mode_amps

    def _set_normal(self):
        """Get normal direction for a 2D monitor. Raise an error if not a 
        2D monitor. """

        if not (isinstance(self.monitor, ModeMonitor) or
            'flux' in self.monitor.store):
            return

        self.norm_ind = np.nonzero(self.monitor.size < fp_eps)[0]
        if self.norm_ind.size !=1:
            log_and_raise(
                f"Exactly one element of the 'size' of Monitor {self.name} "
                "must be zero.", 
                ValueError
            )
            
        self.norm_ind = int(self.norm_ind)

        if isinstance(self.monitor, ModeMonitor):
            self.mode_plane = ModePlane(self.monitor.span, self.norm_ind)

    def _set_tmesh(self, tmesh):
        """ Set the time mesh of the monitor. At the end of a simulation run, 
        the field values are syncronized to be on the ``tmesh`` as defined 
        here. This involves recording one extra step in the solver, and 
        averaging the H-fields, which live on a (n + 1/2)*dt time mesh.
        """
 
        # Step to compare to in order to handle t_start = t_stop
        if np.array(tmesh).size < 1:
            dt = 1e-20
        else:
            dt = tmesh[1] - tmesh[0]

        if self.monitor.t_step is not None:
            self.tind_step = max(1, int(self.monitor.t_step/dt))

        # If t_stop is None, record until the end
        t_stop = self.monitor.t_stop
        if self.monitor.t_stop is None:
            t_stop = tmesh[-1]
            self.tind_end = tmesh.size
        else:
            tend = np.nonzero(tmesh <= self.monitor.t_stop)[0]
            if tend.size > 0:
                self.tind_end = tend[-1] + 1
            else:
                self.tind_end = 0

        # If equal (within dt), record one time step
        if np.abs(self.monitor.t_start - t_stop) < dt:
            self.tind_beg = np.max([self.tind_end - 1, 0])
        else:
            tbeg = np.nonzero(tmesh[0:self.tind_end] >= 
                                self.monitor.t_start)[0]
            if tbeg.size > 0:
                self.tind_beg = tbeg[0]
            else:
                self.tind_beg = self.tind_end

        self.tmesh = tmesh[self.tind_beg:self.tind_end:self.tind_step]
        self.Nt = self.tmesh.size

    def _store_data(self, mnt_data, mnt_inds, field):
        """ Store the raw monitor values returned by the solver.
        
        Parameters
        ----------
        mnt_data : np.ndarray
            An array of shape ``(Np, 3, Nsample)``, where ``Np`` is the total 
            number of monitor points, and ``Nsample`` is either the number of 
            time steps in a ``TimeMonitor``, or the number of requested 
            frequencies in a ``FreqMonitor``.
        mnt_inds : np.ndarray
            An array of shape ``(Np, 3)`` giving the x, y, and z index for 
            each point in the simulation grid.

        Note
        ----
        ``MonitorData.E`` and ``MonitorData.H`` are stored in the format 
        ``(3, indx, indy, indz, samp_ind)``, where ``samp_ind`` is either a 
        time or a frequency index.
        """

        self.mnt_inds = mnt_inds
        self.inds_beg = np.amin(mnt_inds, axis = 0)
        self.inds_end = np.amax(mnt_inds, axis = 0) + 1
        mnt_dims = self.inds_end - self.inds_beg
        if field.lower()=='e':
            self._E = np.zeros((3, mnt_dims[0], mnt_dims[1], mnt_dims[2], 
                        mnt_data.shape[2]), dtype=mnt_data.dtype)
            for ipol in range(3):
                self._E[ipol, mnt_inds[:, 0]-self.inds_beg[0],
                    mnt_inds[:, 1]-self.inds_beg[1],
                    mnt_inds[:, 2]-self.inds_beg[2], :] = mnt_data[:, ipol, :]
        elif field.lower()=='h':
            self._H = np.zeros((3, mnt_dims[0], mnt_dims[1], mnt_dims[2], 
                        mnt_data.shape[2]), dtype=mnt_data.dtype)
            for ipol in range(3):
                self._H[ipol, mnt_inds[:, 0]-self.inds_beg[0],
                    mnt_inds[:, 1]-self.inds_beg[1],
                    mnt_inds[:, 2]-self.inds_beg[2], :] = mnt_data[:, ipol, :]
                    

    def _load_fields(self, mdata, symmetries, Nxyz):
        """ Load the fields returned by the solver. This also applies any 
        symmetries that were present in the simulation. ``mdata`` is an h5py
        group for the corresponding monitor as returned by solver.
        """

        # By default just store index and field values, unless changed below
        self.inds_beg = np.array(mdata["indspan"][0, :])
        self.inds_end = np.array(mdata["indspan"][1, :])
        # Also store copies because inds_beg and inds_end will change later
        inds_beg = np.copy(self.inds_beg)
        inds_end = np.copy(self.inds_end)
        mesh = [np.array(mdata[m]) for m in ['xmesh', 'ymesh', 'zmesh']]
        E = np.array(mdata["E"])
        H = np.array(mdata["H"])
        flux = np.array(mdata["flux"])

        if E.size > 0:
            self._E = E * self.source_norm[None, None, None, None, :]
        else:
            self._E = E
        if H.size > 0:
            self._H = H * self.source_norm[None, None, None, None, :]
        else:
            self._H = H
        if flux.size > 0:
            self._flux = flux * self.source_norm[:, None]**2 
        else:
            self._flux = flux

        # Auxiliary variable for slicing along a given axis
        slices = (slice(None),)*5

        """If symmetries are present, we need to offset the stored fields 
        by half the simulation size in the symmetry direction. Also, if a 
        monitor starts at the symmetry axis, we double its size and 
        pad it with the fields with the correct symmetry eigenvalues. """
        for dim, sym in enumerate(symmetries):

            # Auxiliary variable for symmetry eigenvalue along current axis
            svals = np.ones((3, 1, 1, 1, 1))
            svals[dim] = -1

            if sym==-1:
                self.inds_beg[dim] += Nxyz[dim]//2
                self.inds_end[dim] += Nxyz[dim]//2
                if inds_beg[dim]==0:
                    self.inds_beg[dim] -= inds_end[dim]
                    sl = list(slices)
                    sl[dim+1] = slice(-1, None, -1)
                    # Stack mesh
                    mesh[dim] = np.hstack((mesh[dim][sl[dim + 1]], mesh[dim]))
                    # Stack fields
                    if E.size > 0:
                        self._E = np.concatenate((-svals*self._E[tuple(sl)],
                                    self._E), axis=dim + 1)
                    if H.size > 0:
                        self._H = np.concatenate((svals*self._H[tuple(sl)],
                                    self._H), axis=dim + 1)
                    # Double the flux
                    if flux.size > 0:
                        flux *= 2

            if sym==1:
                # Things are a bit more complicated with PMC symmetry
                self.inds_beg[dim] += Nxyz[dim]//2 - 1
                self.inds_end[dim] += Nxyz[dim]//2 - 1
                if inds_beg[dim]==0:
                    ibeg = 1
                    iend = min(inds_end[dim], Nxyz[dim]//2 + 1)
                    self.inds_beg[dim] -= iend - ibeg - 1
                    self.inds_end[dim] = Nxyz[dim]//2 + iend
                    sl1 = slice(iend - 1, ibeg - 1, -1)
                    sl2 = slice(ibeg, iend)
                    sls1 = list(slices)
                    sls1[dim + 1] = sl1
                    sls2 = list(slices)
                    sls2[dim + 1] = sl2

                    # Stack mesh
                    mesh[dim] = np.hstack((mesh[dim][sl1], mesh[dim][sl2]))
                    # Stack fields
                    if E.size > 0:
                        self._E = np.concatenate((svals*self._E[tuple(sls1)],
                                    self._E[tuple(sls2)]), axis=dim + 1)
                    if H.size > 0:
                        self._H = np.concatenate((-svals*self._H[tuple(sls1)],
                                    self._H[tuple(sls2)]), axis=dim + 1)
                    # Double the flux
                    if flux.size > 0:
                        # This is not exactly correct for PMC.
                        flux *= 2

        self.xmesh = mesh[0]
        self.ymesh = mesh[1]
        self.zmesh = mesh[2]
        self.tmesh = np.array(mdata["tmesh"])
        self.data = True

    def _load_modes(self, mdata):
        """Load the mode amplitude expansion coefficients."""

        if not isinstance(self.monitor, ModeMonitor):
            return

        mplane = self.mode_plane
        ma = np.array(mdata["mode_amps"])
        self._mode_amps = ma * self.source_norm[None, :, None]

        try:
            modes_E = np.array(mdata["modes_E"])
            modes_H = np.array(mdata["modes_H"])
            modes_n = np.array(mdata["modes_n"])
            modes_k = np.array(mdata["modes_k"])
            Nf, Nm = modes_n.shape
            
            for find in range(Nf):
                mplane.modes[find] = []
                for mind in range(Nm):
                    E = modes_E[find, mind, :, :, :]
                    H = modes_H[find, mind, :, :, :]
                    n = modes_n[find, mind]
                    k = modes_k[find, mind]
                    mplane.modes[find].append(Mode(E, H, n, k))
        except:
            pass

    def set_source_norm(self, src_data):
        """Normalize the stored fields.
        
        Parameters
        ----------
        src_data : src_data or None
            A :class:`.SourceData` object from whose spectrum the fields are 
            normalized.
        """

        if isinstance(self.monitor, TimeMonitor):
            self.source_norm = np.ones((1, ), dtype=float_)
        elif src_data is None:
            self.source_norm = np.ones((len(self.freqs), ), dtype=float_)
        else:
            spectrum = src_data._get_spectrum(self.freqs)
            self.source_norm = np.array(1/np.abs(spectrum))